@@ -12,11 +12,13 @@ import urllib.parse
 import requests
 import ast
 import numpy
-
+from unittest.case import TestCase
 from unitgrade.runners import UTextResult
 from unitgrade.utils import gprint, Capturing2, Capturing
-colorama.init(autoreset=True)  # auto resets your settings after every output
+from unitgrade.artifacts import StdCapturing
 
+
+colorama.init(autoreset=True)  # auto resets your settings after every output
 numpy.seterr(all='raise')
 
 def setup_dir_by_class(C, base_dir):
@@ -24,6 +26,53 @@ def setup_dir_by_class(C, base_dir):
     return base_dir, name
 
 
+_DASHBOARD_COMPLETED_MESSAGE = "Dashboard> Evaluation completed."
+
+# Consolidate this code.
+class classmethod_dashboard(classmethod):
+    def __init__(self, f):
+        def dashboard_wrap(cls: UTestCase):
+            if not cls._generate_artifacts:
+                f(cls)
+                return
+            from unitgrade.utils import DKPupDB
+            db = DKPupDB(cls._artifact_file_for_setUpClass())
+            r = np.random.randint(1000 * 1000)
+            db.set('run_id', r)
+            db.set('coverage_files_changed', None)
+
+            state_ = 'fail'
+            try:
+                _stdout = sys.stdout
+                _stderr = sys.stderr
+                std_capture = StdCapturing(stdout=sys.stdout, stderr=sys.stderr, db=db, mute=False)
+
+                # Run this unittest and record all of the output.
+                # This is probably where we should hijack the stdout output and save it -- after all, this is where the test is actually run.
+                # sys.stdout = stdout_capture
+                sys.stderr = std_capture.dummy_stderr
+                sys.stdout = std_capture.dummy_stdout
+                db.set("state", "running")
+                f(cls)
+                state_ = 'pass'
+            except Exception as e:
+                from werkzeug.debug.tbtools import DebugTraceback, _process_traceback
+                state_ = 'fail'
+                db.set('state', state_)
+                exi = e
+                dbt = DebugTraceback(exi)
+                sys.stderr.write(dbt.render_traceback_text())
+                html = dbt.render_traceback_html(include_title="hello world")
+                db.set('wz_stacktrace', html)
+                raise e
+            finally:
+                db.set('state', state_)
+                std_capture.dummy_stdout.write_mute(_DASHBOARD_COMPLETED_MESSAGE)
+                sys.stdout = _stdout
+                sys.stderr = _stderr
+                std_capture.close()
+        super().__init__(dashboard_wrap)
+
 class Report:
     title = "report title"
     abbreviate_questions = False # Should the test items start with 'Question ...' or just be q1).
@@ -34,7 +83,7 @@ class Report:
     pack_imports = []
     individual_imports = []
 
-    _remote_check_cooldown_seconds = 1 # Seconds between remote check of report.
+    _remote_check_cooldown_seconds = 1  # Seconds between remote check of report.
     nL = 120  # Maximum line width
     _config = None  # Private variable. Used when collecting results from student computers. Should only be read/written by teacher and never used for regular evaluation.
     _setup_mode = False # True if test is being run in setup-mode, i.e. will not fail because of bad configurations, etc.
@@ -52,6 +101,11 @@ class Report:
     def _file(self):
         return inspect.getfile(type(self))
 
+    def _artifact_file(self):
+        """ File for the artifacts DB (thread safe). This file is optinal. Note that it is a pupdb database file.
+        Note the file is shared between all sub-questions. """
+        return os.path.join(os.path.dirname(self._file()), "unitgrade_data/main_config_"+ os.path.basename(self._file()[:-3]) + ".artifacts.pkl")
+
     def _is_run_in_grade_mode(self):
         """ True if this report is being run as part of a grade run. """
         return self._file().endswith("_grade.py") # Not sure I love this convention.
@@ -66,7 +120,6 @@ class Report:
         relative_path = os.path.relpath(self._file(), root_dir)
         modules = os.path.normpath(relative_path[:-3]).split(os.sep)
         relative_path = relative_path.replace("\\", "/")
-
         return root_dir, relative_path, modules
 
     def __init__(self, strict=False, payload=None):
@@ -96,9 +149,10 @@ class Report:
                 q._report = self
         for q, _ in self.questions:
             q._setup_answers_mode = True
+            # q._generate_artifacts = False # Disable artifact generation when the report is being set up.
 
         from unitgrade import evaluate_report_student
-        evaluate_report_student(self, unmute=verbose, noprogress=not verbose)
+        evaluate_report_student(self, unmute=verbose, noprogress=not verbose, generate_artifacts=False) # Disable artifact generation.
 
         # self.main()  # Run all tests in class just to get that out of the way...
         report_cache = {}
@@ -106,14 +160,17 @@ class Report:
             # print(self.questions)
             if hasattr(q, '_save_cache'):
                 q()._save_cache()
-                print("q is", q())
-                # q()._cache_put('time', q.time) # = q.time
+                # print("q is", q())
                 report_cache[q.__qualname__] = q._cache2
             else:
                 report_cache[q.__qualname__] = {'no cache see _setup_answers in framework.py': True}
         if with_coverage:
             for q, _ in self.questions:
                 q._with_coverage = False
+
+        # report_cache is saved on a per-question basis.
+        # it could also contain additional information such as runtime metadata etc. This may not be appropriate to store with the invidivual questions(?).
+        # In this case, the function should be re-defined.
         return report_cache
 
     def set_payload(self, payloads, strict=False):
@@ -128,15 +185,11 @@ class Report:
         if not url.endswith("/"):
             url += "/"
         snapshot_file = os.path.dirname(self._file()) + "/unitgrade_data/.snapshot"
-        # print("Sanity checking time using snapshot", snapshot_file)
-        # print("and using self-identified file", self._file())
-
         if os.path.isfile(snapshot_file):
             with open(snapshot_file, 'r') as f:
                 t = f.read()
                 if (time.time() - float(t)) < self._remote_check_cooldown_seconds:
                     return
-        # print("Is this file run in local mode?", self._is_run_in_grade_mode())
 
         if self.url.startswith("https://gitlab"):
             # Try to turn url into a 'raw' format.
@@ -163,18 +216,6 @@ class Report:
                     print("You can find the most recent code here:")
                     print(self.url)
                     raise Exception(f"Version of grade script does not match the remote version. Please update using git pull")
-                #
-                # # node = ast.parse(text)
-                # # classes = [n for n in node.body if isinstance(n, ast.ClassDef) if n.name == self.__class__.__name__][0]
-                #
-                # # for b in classes.body:
-                # #     print(b.)
-                #     # if b.targets[0].id == "version":
-                #         # print(b)
-                #         # print(b.value)
-                #         version_remote = b.value.value
-                #         break
-                # if version_remote != self.version:
             else:
                 text = requests.get(raw_url).text
                 node = ast.parse(text)
@@ -198,7 +239,7 @@ class Report:
 
                 for (q,_) in self.questions:
                     qq = q(skip_remote_check=True)
-                    cfile = qq._cache_file()
+                    cfile = q._cache_file()
 
                     relpath = os.path.relpath(cfile, os.path.dirname(self._file()))
                     relpath = relpath.replace("\\", "/")
@@ -226,6 +267,13 @@ class Report:
                         f.write(f"{t}")
 
 def get_hints(ss):
+    """ Extract all blocks of the forms:
+
+    Hints:
+    bla-bla.
+
+    and returns the content unaltered.
+    """
     if ss == None:
         return None
     try:
@@ -240,19 +288,24 @@ def get_hints(ss):
         ss = ss[ss.lower().find(h) + len(h) + 1:]
         ss = "\n".join([l for l in ss.split("\n") if not l.strip().startswith(":")])
         ss = textwrap.dedent(ss).strip()
+        # if ss.startswith('*'):
+        #     ss = ss[1:].strip()
         return ss
     except Exception as e:
         print("bad hints", ss, e)
 
 
 class UTestCase(unittest.TestCase):
+    # a = 234
     _outcome = None  # A dictionary which stores the user-computed outcomes of all the tests. This differs from the cache.
     _cache = None  # Read-only cache. Ensures method always produce same result.
     _cache2 = None  # User-written cache.
     _with_coverage = False
     _covcache = None # Coverage cache. Written to if _with_coverage is true.
     _report = None  # The report used. This is very, very hacky and should always be None. Don't rely on it!
+    _run_in_report_mode = True
 
+    _generate_artifacts = True # Whether the file will generate the artifact .json files. This is used in the _grade-script mode.
     # If true, the tests will not fail when cache is used. This is necesary since otherwise the cache will not be updated
     # during setup, and the deploy script must be run many times.
     _setup_answers_mode = False
@@ -261,8 +314,6 @@ class UTestCase(unittest.TestCase):
         if hasattr(self, '_stdout') and self._stdout is not None:
             file = self._stdout
         else:
-            # self._stdout = sys.stdout
-            # sys._stdout = io.StringIO()
             file = sys.stdout
         return Capturing2(stdout=file)
 
@@ -275,13 +326,70 @@ class UTestCase(unittest.TestCase):
                 return title
         return cls.__qualname__
 
-    @classmethod
-    def reset(cls):
-        print("Warning, I am not sure UTestCase.reset() is needed anymore and it seems very hacky.")
-        raise Exception("reset called in test case. This method is deprecated.")
-        cls._outcome = None
-        cls._cache = None
-        cls._cache2 = None
+    def run(self, result):
+        # print("Run called in test framework...", self._generate_artifacts)
+        if not self._generate_artifacts:
+            return super().run(result)
+        from unitgrade.artifacts import StdCapturing
+        from unitgrade.utils import DKPupDB
+
+        db = DKPupDB(self._artifact_file())
+        db.set("state", "running")
+        db.set('run_id', np.random.randint(1000*1000))
+        db.set('coverage_files_changed', None)
+
+
+        _stdout = sys.stdout
+        _stderr = sys.stderr
+
+        std_capture = StdCapturing(stdout=sys.stdout, stderr=sys.stderr, db=db, mute=False)
+
+        # stderr_capture = StdCapturing(sys.stderr, db=db)
+        # std_err_capture = StdCapturing(sys.stderr, "stderr", db=db)
+        state_ = None
+        try:
+            # Run this unittest and record all of the output.
+            # This is probably where we should hijack the stdout output and save it -- after all, this is where the test is actually run.
+            # sys.stdout = stdout_capture
+            sys.stderr = std_capture.dummy_stderr
+            sys.stdout = std_capture.dummy_stdout
+
+            result_ = TestCase.run(self, result)
+
+            from werkzeug.debug.tbtools import DebugTraceback, _process_traceback
+            # print(result_._excinfo[0])
+            actual_errors = []
+            for test, err in self._error_fed_during_run:
+                if err is None:
+                    continue
+                else:
+                    import traceback
+                    # traceback.print_tb(err[2])
+                    actual_errors.append(err)
+
+            if len(actual_errors) > 0:
+                ex, exi, tb = actual_errors[0]
+                exi.__traceback__ = tb
+                dbt = DebugTraceback(exi)
+                sys.stderr.write(dbt.render_traceback_text())
+                html = dbt.render_traceback_html(include_title="hello world")
+                db.set('wz_stacktrace', html)
+                # db.set('state', 'fail')
+                state_ = "fail"
+            else:
+                state_ = "pass"
+        except Exception as e:
+            state_ = "fail"
+            import traceback
+            traceback.print_exc()
+            raise e
+        finally:
+            db.set('state', state_)
+            std_capture.dummy_stdout.write_mute(_DASHBOARD_COMPLETED_MESSAGE)
+            sys.stdout = _stdout
+            sys.stderr = _stderr
+            std_capture.close()
+        return result_
 
     def _callSetUp(self):
         if self._with_coverage:
@@ -300,26 +408,16 @@ class UTestCase(unittest.TestCase):
             from pathlib import Path
             from snipper import snipper_main
             try:
-                # print("Stoppping coverage...")
                 self.cov.stop()
-                # print("Coverage was stopped")
-                # self.cov.html_report()
-                # print("Success!")
             except Exception as e:
                 print("Something went wrong while tearing down coverage test")
                 print(e)
             data = self.cov.get_data()
             base, _, _ = self._report._import_base_relative()
-            # print("Measured coverage files", data.measured_files)
             for file in data.measured_files():
-                # print(file)
                 file = os.path.normpath(file)
                 root = Path(base)
                 child = Path(file)
-                # print("root", root, "child", child)
-                # print(child, "is in parent?", root in child.parents)
-                # print(child.parents)
-
                 if root in child.parents:
                     # print("Reading file", child)
                     with open(child, 'r') as f:
@@ -331,31 +429,14 @@ class UTestCase(unittest.TestCase):
                     if len(lines) != len(lines2):
                         for k in range(len(lines)):
                             print(k, ">", lines[k], "::::::::", lines2[k])
-
-                        # print("-" * 100)
-                        # print("\n".join(lines))
-                        # print("-"*100)
-                        # print("\n".join(lines2))
-                        # print("-" * 100)
                         print("Snipper failure; line lenghts do not agree. Exiting..")
                         print(child, "len(lines) == len(lines2)", len(lines), len(lines2))
                         import sys
                         sys.exit()
 
                     assert len(lines) == len(lines2)
-                    # print("In file ", file, "context by lineno", data.contexts_by_lineno(file))
                     for ll in data.contexts_by_lineno(file):
-                        # For empty files (e.g. __init__) there is a potential bug where coverage will return the file but lines2 will be = [].
-                        # print("loop B: ll is", ll)
                         l = ll-1
-                        # print(l)
-                        # l1 = (lines[l] + " "*1000)[:80]
-                        # l2 = (lines2[l] + " "*1000)[:80]
-                        # print("l is", l, l1, " " + l2, "file", file)
-                        # print("Checking if statement: ")
-                        # print(l, lines2)
-                        # print(">> ", lines2[l])
-                        # print(">", lines2[l].strip(), garb)
                         if l < len(lines2) and lines2[l].strip() == garb:
                             # print("Got a hit at l", l)
                             rel = os.path.relpath(child, root)
@@ -394,7 +475,7 @@ class UTestCase(unittest.TestCase):
         self._cache_put((self.cache_id(), 'title'), value)
 
     def _get_outcome(self):
-        if not (self.__class__, '_outcome') or self.__class__._outcome is None:
+        if not hasattr(self.__class__, '_outcome') or self.__class__._outcome is None:
             self.__class__._outcome = {}
         return self.__class__._outcome
 
@@ -425,9 +506,6 @@ class UTestCase(unittest.TestCase):
         # Find the report class this class is defined within.
         if skip_remote_check:
             return
-        import inspect
-
-        # file = inspect.getfile(self.__class__)
         import importlib, inspect
         found_reports = []
         # print("But do I have report", self._report)
@@ -449,11 +527,6 @@ class UTestCase(unittest.TestCase):
             report = found_reports[0]
             report()._check_remote_versions()
 
-        # self._get_report_class()
-
-
-    # def _get_report_class(self):
-    #     pass
 
     def _ensure_cache_exists(self):
         if not hasattr(self.__class__, '_cache') or self.__class__._cache == None:
@@ -484,8 +557,7 @@ class UTestCase(unittest.TestCase):
         key = (self.cache_id(), 'assert')
         if not self._cache_contains(key):
             print("Warning, framework missing", key)
-            self.__class__._cache[
-                key] = {}  # A new dict. We manually insert it because we have to use that the dict is mutable.
+            self.__class__._cache[key] = {}  # A new dict. We manually insert it because we have to use that the dict is mutable.
         cache = self._cache_get(key)
         id = self._assert_cache_index
         _expected = cache.get(id, f"Key {id} not found in cache; framework files missing. Please run deploy()")
@@ -527,7 +599,6 @@ class UTestCase(unittest.TestCase):
         diff = np.abs(a1 - a2)
         return diff
 
-
     def assertLinf(self, first, second=None, tol=1e-5, msg=None):
         """ Test in the L_infinity norm.
         :param first:
@@ -577,12 +648,25 @@ class UTestCase(unittest.TestCase):
                 self.assertFalse(max_diff >= tol, msg=f'Input arrays are not equal within tolerance {tol}')
                 # self.assertEqual(first, second, msg=msg + f"Not equal within tolerance {tol}")
 
-    def _cache_file(self):
-        return os.path.dirname(inspect.getabsfile(type(self))) + "/unitgrade_data/" + self.__class__.__name__ + ".pkl"
+    @classmethod
+    def _cache_file(cls):
+        return os.path.dirname(inspect.getabsfile(cls)) + "/unitgrade_data/" + cls.__name__ + ".pkl"
+
+    @classmethod
+    def _artifact_file_for_setUpClass(cls):
+        file = os.path.join(os.path.dirname(cls._cache_file()), ""+cls.__name__+"-setUpClass.json")
+        print("_artifact_file_for_setUpClass(cls): will return", file, "__class__", cls)
+        # cf = os.path.dirname(inspect.getabsfile(cls)) + "/unitgrade_data/" + cls.__name__
+        return file
+
+    def _artifact_file(self):
+        """ File for the artifacts DB (thread safe). This file is optinal. Note that it is a pupdb database file.
+        Note the file is shared between all sub-questions. """
+        return os.path.join(os.path.dirname(self.__class__._cache_file()), '-'.join(self.cache_id()) + ".json")
 
     def _save_cache(self):
         # get the class name (i.e. what to save to).
-        cfile = self._cache_file()
+        cfile = self.__class__._cache_file()
         if not os.path.isdir(os.path.dirname(cfile)):
             os.makedirs(os.path.dirname(cfile))
 
@@ -595,7 +679,7 @@ class UTestCase(unittest.TestCase):
         if self._cache is not None:  # Cache already loaded. We will not load it twice.
             return
             # raise Exception("Loaded cache which was already set. What is going on?!")
-        cfile = self._cache_file()
+        cfile = self.__class__._cache_file()
         if os.path.exists(cfile):
             try:
                 with open(cfile, 'rb') as f:
@@ -607,8 +691,55 @@ class UTestCase(unittest.TestCase):
         else:
             print("Warning! data file not found", cfile)
 
+    def _get_coverage_files(self):
+        key = (self.cache_id(), 'coverage')
+        # CC = None
+        # if self._cache_contains(key):
+        return self._cache_get(key, []) # Anything wrong with the empty list?
+        # return CC
+
+    def _get_hints(self):
+        """
+            This code is run when the test is set up to generate the hints and store them in an artifact file. It may be beneficial to simple compute them beforehand
+            and store them in the local unitgrade pickle file. This code is therefore expected to superceede the alterative code later.
+        """
+        hints = []
+        # print("Getting hint")
+        key = (self.cache_id(), 'coverage')
+        if self._cache_contains(key):
+            CC = self._cache_get(key)
+            # cl, m = self.cache_id()
+            # print("Getting hint using", CC)
+            # Insert newline to get better formatting.
+            # gprint(
+            #     f"\n> An error occured during the test: {cl}.{m}. The following files/methods has code in them you are supposed to edit and may therefore be the cause of the problem:")
+            for file in CC:
+                rec = CC[file]
+                # gprint(f">   * {file}")
+                for l in rec:
+                    _, comments = CC[file][l]
+                    hint = get_hints(comments)
+
+                    if hint != None:
+                        hints.append((hint, file, l))
+
+        doc = self._testMethodDoc
+        # print("doc", doc)
+        if doc is not None:
+            hint = get_hints(self._testMethodDoc)
+            if hint is not None:
+                hints = [(hint, None, self.cache_id()[1])] + hints
+
+        return hints
+
     def _feedErrorsToResult(self, result, errors):
-        """ Use this to show hints on test failure. """
+        """ Use this to show hints on test failure.
+        It feeds error to the result -- so if there are errors, they will crop up here
+        """
+        self._error_fed_during_run = errors.copy() # import to copy the error list.
+
+        # result._test._error_fed_during_run = errors.copy()
+
         if not isinstance(result, UTextResult):
             er = [e for e, v in errors if v != None]
             # print("Errors are", errors)
@@ -653,8 +784,9 @@ class UTestCase(unittest.TestCase):
                         except Exception as e:
                             print("Bad stuff in hints. ")
                             print(hints)
-
+        # result._last_errors = errors
         super()._feedErrorsToResult(result, errors)
+        b = 234
 
     def startTestRun(self):
         super().startTestRun()
@@ -684,4 +816,5 @@ class NotebookTestCase(UTestCase):
 
     @property
     def nb(self):
-        return self.__class__._nb
\ No newline at end of file
+        return self.__class__._nb
+ # 870.
\ No newline at end of file
